/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/* SCCS ID: @(#)UnixStandardize.c	1.9 2/16/94 */
/*************************************************************************
 * FILE:      UnixStandardize.c
 * Author:    Claude Saunders
 * 
 * Purpose:   Program for standardizing magnets. An SDDS data file is
 *            read which specifies which devices (or PV's) are to be
 *            standardized and how. Subroutine records on IOC's are
 *            configured appropriately and processed to begin standard-
 *            ization. Various options allow the user to start/stop/query
 *            standardization.
 *
 * Mods:      login mm/dd/yy: description
 *************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <stddef.h>
#include <cadef.h>
#include <SDDS.h>          /* SDDS library header */

#define CA_TIMEOUT 10.0

static int debug = 0;      /* Internal debugging flag */

/* Function Prototypes */
void sleep_us(unsigned int);
int validate_time(long min, long sec);
int process_cmdline_args(int argc, char *argv[], char *spec, int *query, 
			 int *configure, int *stop, double *pendIoTime);
void usage(char *name);

static struct sub_rec_pvs {
  char sub_proc[255];        /* for initiating standardization */
  chid sub_proc_chid;
  char sub_snam[255];        /* is proper subroutine attached ? */
  chid sub_snam_chid;
  char sub_b[255];           /* arguments to standardization in fields b-h */
  chid sub_b_chid;
  char sub_c[255];
  chid sub_c_chid;
  char sub_d[255];
  chid sub_d_chid;
  char sub_e[255];
  chid sub_e_chid;
  char sub_f[255];
  chid sub_f_chid;
  char sub_g[255];
  chid sub_g_chid;
  char sub_h[255];
  chid sub_h_chid;
};
  
int main(int argc, char *argv[]) {
  SDDS_TABLE table;
  char **name;        /* device name or process variable */
  char **type;        /* dev or pv */
  double *from;       /* start point of conditioning cycle */
  double *to;         /* end point of conditioning cycle */
  long *min, *sec;    /* time for transition from..to   */
  long *ncycles;      /* # of cycles in standardization */
  double *finish;     /* final setpoint to stop at */
  char **approach;    /* approach finish from "above" or "below" */
  double *dblapp;     /* same as approach but 0.0 for below 1.0 for above */
  char **pvname;      /* name of subroutine record process variable */

  struct sub_rec_pvs *sub;
  int query, configure, stop;
  double pendIoTime;
  char specfilename[255];     /* specification file name */
  char *text, *contents;      /* text and contents fields of SDDS descrip. */
  long nrows;
  int status;
  int exitstatus = 0;
  long i, j, k, l, nsupplies;
  double *state;       /* gets proc field value from subroutine record */
  int *ndevs;          /* number of atomic devices in device i */
  char **subname;
  char **dname;        /* names of individual atomic devices */
  short start = 1;
  short halt = 0;
  char buffer[255];

  fprintf(stdout,"Magnet Standardization Utility\n");
  fprintf(stdout,"------------------------------\n");

  process_cmdline_args(argc, argv, specfilename, &query, &configure, &stop, &pendIoTime);

  if (!SDDS_InitializeInput(&table, specfilename)) {
    fprintf(stderr,"%s: SDDS error: unable to initialize with file %s\n",
	    argv[0], specfilename);
    exit(1);
  }
  if (SDDS_ReadTable(&table) <= 0) {
    fprintf(stderr,"%s: SDDS error: unable to read data table from %s\n",
	    argv[0], specfilename);
    exit(1);
  }
  SDDS_SetColumnFlags(&table, 1);
  SDDS_SetRowFlags(&table, 1);

  contents = NULL;
  /*
  if (!SDDS_GetDescription(&table, NULL, &contents)) {
    fprintf(stderr,"%s: SDDS error: unable to get description from file %s\n",
	    argv[0], specfilename);
    exit(1);
  }
  if (contents==NULL || strcmp(contents,"standardization specification")) {
    fprintf(stderr,"%s: SDDS error: file %s is not a standardization specification\n",argv[0],specfilename);
    exit(1);
  }
  */

  if ((nrows = SDDS_CountRowsOfInterest(&table)) == -1) {
    fprintf(stderr,"%s: SDDS error: unable to get row count\n",argv[0]);
    exit(1);
  }

  if ((name = SDDS_GetColumn(&table, "ControlName"))==NULL) {
    fprintf(stderr,"%s: SDDS error: unable to read controlname column\n",argv[0]);
    exit(1);
  }
  if ((type = SDDS_GetColumn(&table, "ControlType"))==NULL) {
    fprintf(stderr,"%s: SDDS error: unable to read controltype column\n",argv[0]);
    exit(1);
  }
  if ((from = SDDS_GetColumn(&table, "From"))==NULL) {
    fprintf(stderr,"%s: SDDS error: unable to read from column\n",argv[0]);
    exit(1);
  }
  if ((to = SDDS_GetColumn(&table, "To"))==NULL) {
    fprintf(stderr,"%s: SDDS error: unable to read to column\n",argv[0]);
    exit(1);
  }
  if ((min = SDDS_GetColumn(&table, "Minutes"))==NULL) {
    fprintf(stderr,"%s: SDDS error: unable to read min column\n",argv[0]);
    exit(1);
  }
  if ((sec = SDDS_GetColumn(&table, "Seconds"))==NULL) {
    fprintf(stderr,"%s: SDDS error: unable to read sec column\n",argv[0]);
    exit(1);
  }
  if ((ncycles = SDDS_GetColumn(&table, "NumCycles"))==NULL) {
    fprintf(stderr,"%s: SDDS error: unable to read ncycles column\n",argv[0]);
    exit(1);
  }
  if ((finish = SDDS_GetColumn(&table, "Finish"))==NULL) {
    fprintf(stderr,"%s: SDDS error: unable to read finish column\n",argv[0]);
    exit(1);
  }
  if ((approach = SDDS_GetColumn(&table, "Approach"))==NULL) {
    fprintf(stderr,"%s: SDDS error: unable to read approach column\n",argv[0]);
    exit(1);
  }

  for (i=0 ; i < nrows ; i++)
    if (validate_time(min[i],sec[i])) {
      fprintf(stderr,"%s: %s has invalid time specified\n",argv[0],name[i]);
      exit(1);
    }

  ndevs = (int *)malloc(sizeof(int)*nrows);

  for (i=0, nsupplies=0 ; i < nrows ; i++) {
    if ((!strcmp(type[i], "dev")) || (!strcmp(type[i], "pv"))) {
      nsupplies += 1;
      ndevs[i] = 1;
    }
    else {
      fprintf(stderr,"%s: illegal ControlType field for %s\n",
	      argv[0],name[i]);
      fprintf(stderr,"%s: exiting, no standardization initiated\n");
      exit(1);
    }
  }
  
  sub = (struct sub_rec_pvs *)malloc(sizeof(struct sub_rec_pvs)*nsupplies);
  dblapp = (double *)malloc(sizeof(double)*nsupplies);
  state = (double *)malloc(sizeof(double)*nsupplies);
  pvname = (char **)malloc(sizeof(char *)*nsupplies);
  subname = (char **)malloc(sizeof(char *)*nsupplies);
  dname = (char **)malloc(sizeof(char *)*nsupplies);
  for (i=0 ; i < nsupplies ; i++) {
    pvname[i] = (char *)malloc(sizeof(char)*255);
    subname[i] = (char *)malloc(sizeof(char)*255);
    dname[i] = (char *)malloc(sizeof(char)*255);
  }
  
  for (i=0 ; i < nrows ; i++)
    dblapp[i] = (strcmp(approach[i],"above"))?0.0:1.0;

  for (i=0 ; i < nrows ; i++) {
    strcpy(dname[i], name[i]);
  }
    
  /* Build pv names for needed sub record fields based on given pv name */
  for (i=0 ; i < nrows ; i++) {
    if (!strcmp(type[i],"dev")) {        /* device name specified in input */
      strcpy(buffer, name[i]);
      strcat(buffer, ":StandardizeSUB");
      strcpy(pvname[i], buffer);
    } else {
      strcpy(pvname[i], name[i]);
    }
  }

  for (i=0, k=0 ; i < nrows ; i++) {           /* for each row in sdds file */
    for (l=0 ; l < ndevs[i] ; l++, k++) {      /* for each atomic in comp */
      strcpy(sub[k].sub_proc, pvname[k]);
      strcat(sub[k].sub_proc, ".PROC");
      strcpy(sub[k].sub_snam, pvname[k]);
      strcat(sub[k].sub_snam, ".SNAM");
      strcpy(sub[k].sub_b, pvname[k]);
      strcat(sub[k].sub_b, ".B");
      strcpy(sub[k].sub_c, pvname[k]);
      strcat(sub[k].sub_c, ".C");
      strcpy(sub[k].sub_d, pvname[k]);
      strcat(sub[k].sub_d, ".D");
      strcpy(sub[k].sub_e, pvname[k]);
      strcat(sub[k].sub_e, ".E");
      strcpy(sub[k].sub_f, pvname[k]);
      strcat(sub[k].sub_f, ".F");
      strcpy(sub[k].sub_g, pvname[k]);
      strcat(sub[k].sub_g, ".G");
      strcpy(sub[k].sub_h, pvname[k]);
      strcat(sub[k].sub_h, ".H");
    }
  }

  fprintf(stdout,"Searching for devices..."); fflush(stdout);
  for (i=0, k=0 ; i < nrows ; i++) {
    for (l=0 ; l < ndevs[i] ; l++, k++) {
      status = ca_search(sub[k].sub_proc, &(sub[k].sub_proc_chid));
      if (status != ECA_NORMAL) {
	fprintf(stderr,"%s: unable to connect with %s\n",argv[0],dname[k]);
	exit(1);
      }
      status = ca_search(sub[k].sub_snam, &(sub[k].sub_snam_chid));
      status = ca_search(sub[k].sub_b, &(sub[k].sub_b_chid));
      status |= ca_search(sub[k].sub_c, &(sub[k].sub_c_chid));
      status |= ca_search(sub[k].sub_d, &(sub[k].sub_d_chid));
      status |= ca_search(sub[k].sub_e, &(sub[k].sub_e_chid));
      status |= ca_search(sub[k].sub_f, &(sub[k].sub_f_chid));
      status |= ca_search(sub[k].sub_g, &(sub[k].sub_g_chid));
      status |= ca_search(sub[k].sub_h, &(sub[k].sub_h_chid));
      if (status != ECA_NORMAL) {
	fprintf(stderr,"%s: unable to connect with %s\n",argv[0],dname[k]);
	exit(1);
      }
    }
  }
  status = ca_pend_io(pendIoTime);
  if (status != ECA_NORMAL) {
    fprintf(stderr,"%s: unable to complete connections\n",argv[0]);
    exit(1);
  }
  fprintf(stdout,"done\n");

  /* Check to make sure subroutine is standardize, not degauss */
  for (i=0, k=0 ; i < nrows ; i++) {
    for (l=0 ; l < ndevs[i] ; l++, k++) {
      status = ca_get(DBR_STRING, sub[k].sub_snam_chid, subname[k]);
      if (status != ECA_NORMAL) {
	fprintf(stderr,"%s: unable to ca_get from %s\n",argv[0],dname[k]);
	exit(1);
      }
      status = ca_pend_io(pendIoTime);
      if (status != ECA_NORMAL) {
	fprintf(stderr,"%s: unable to flush ca_get\n",argv[0]);
	exit(1);
      }
      if (strcmp(subname[k],"standardizeProc")) {
	fprintf(stderr,"%25s : device has wrong subroutine specified for standardization\n",dname[k]);
	fprintf(stderr,"%s: exiting, no standardization performed\n",argv[0]);
	exit(1);
      }
    }
  }


  if (stop) {
    fprintf(stdout,"Halting standardization..."); fflush(stdout);
    for (i=0, k=0 ; i < nrows ; i++) {
      for (l=0 ; l < ndevs[i] ; l++, k++) {
	status = ca_put(DBR_SHORT, sub[k].sub_proc_chid, &halt);
	if (status != ECA_NORMAL) {
	  fprintf(stderr,"%s: unable to ca_put to PROC field\n",argv[0]);
	}
	status = ca_pend_io(pendIoTime);
	if (status != ECA_NORMAL) {
	  fprintf(stderr,"%s: unable to complete ca_put to PROC field\n",
		  argv[0]);
	}
      }
    }
    fprintf(stdout,"done\n");
  }

  for (i=0, k=0 ; i < nrows ; i++) {
    for (l=0 ; l < ndevs[i] ; l++, k++) {
      status = ca_get(DBR_DOUBLE, sub[k].sub_proc_chid, &(state[k]));
      if (status != ECA_NORMAL) {
	fprintf(stderr,"%s: unable to ca_get from %s\n",argv[0],dname[k]);
	exit(1);
      }
      status = ca_pend_io(pendIoTime);
      if (status != ECA_NORMAL) {
	fprintf(stderr,"%s: unable to flush ca_get\n",argv[0]);
	exit(1);
      }
      if (!stop) {
	if (state[k] != 0.0) {
	  fprintf(stdout,"%25s : standardizing active\n",dname[k]);
	}
	else {
	  fprintf(stdout,"%25s : standardizing inactive\n",dname[k]);
	}
      }
    }
  }
    
  if (!query && !stop) {
    fprintf(stdout,"Configure records for standardization..."); fflush(stdout);
    for (i=0, k=0 ; i < nrows ; i++) {
      for (l=0 ; l < ndevs[i] ; l++, k++) {
	if (state[k]==0.0) {
	  status = ca_put(DBR_DOUBLE, sub[k].sub_b_chid, &(from[i]));
	  status |= ca_put(DBR_DOUBLE, sub[k].sub_c_chid, &(to[i]));
	  status |= ca_put(DBR_LONG, sub[k].sub_d_chid, &(min[i]));
	  status |= ca_put(DBR_LONG, sub[k].sub_e_chid, &(sec[i]));
	  status |= ca_put(DBR_LONG, sub[k].sub_f_chid, &(ncycles[i]));
	  status |= ca_put(DBR_DOUBLE, sub[k].sub_g_chid, &(finish[i]));
	  status |= ca_put(DBR_DOUBLE, sub[k].sub_h_chid, &(dblapp[i]));
	  if (status != ECA_NORMAL) {
	    fprintf(stderr,"%s: unable to ca_put to %s\n",argv[0],dname[k]);
	    exit(1);
	  }
	}
      }
      status = ca_pend_io(pendIoTime);
      if (status != ECA_NORMAL) {
	fprintf(stderr,"%s: unable to complete ca_puts\n",argv[0]);
	exit(1);
      }
    }
    fprintf(stdout,"done\n");
  }
  
  if (!query && !stop && !configure) {
    fprintf(stdout,"Initiating standardization..."); fflush(stdout);
    sleep(1);
    for (i=0, k=0 ; i < nrows ; i++) {
      for (l=0 ; l < ndevs[i] ; l++, k++) {
	if (state[k]==0.0) {
	  status = ca_put(DBR_SHORT, sub[k].sub_proc_chid, &start);
	  if (status != ECA_NORMAL) {
	    fprintf(stderr,"%s: unable to ca_put to PROC field\n",argv[0]);
	    exit(1);
	  }
	  status = ca_pend_io(pendIoTime);
	  if (status != ECA_NORMAL) {
	    fprintf(stderr,"%s: unable to complete ca_put to PROC field\n",argv[0]);
	    exit(1);
	  }
	}
	sleep_us(100000);
      }
    }
    fprintf(stdout,"done\n");
  }
  /*  exit(0);*/
  ca_task_exit();
  return(0);
}

/*************************************************************************
 * FUNCTION : validate_time()
 * PURPOSE  : Makes sure that time specs are valid. Ie. minutes and
 *            seconds are between 0 and 60.
 * ARGS in  : min:   minutes value for device i
 *            sec:   seconds value for device i
 * ARGS out : none
 * GLOBAL   : nothing
 * RETURNS  : 0 - time values valid 
 *            -1 - at least one time value is invalid
 ************************************************************************/
int validate_time(long min, long sec) {
  if (min < 0 || min > 60 || sec < 0 || sec > 60)
    return(-1);
  else
    return(0);
}

/*************************************************************************
 * FUNCTION : process_cmdline_args()
 * PURPOSE  : Get command line arguments and set flags appropriately.
 * ARGS in  : argc, argv - standard unix
 * ARGS out : spec, query, configure, stop
 * GLOBAL   : nothing
 * RETURNS  : 0
 ************************************************************************/  
int
process_cmdline_args(int argc, char *argv[], char *spec, int *query, 
		     int *configure, int *stop, double *pendIoTime)
{
  int i, dd;
  int cmderror = 0;
  /*  void usage();*/

  /* Process command line options */
  if (argc < 2 || argc > 5)
    cmderror++;
  else {
    *query = 0;
    *configure = 0;
    *stop = 0;
    *spec = '\0';
    *pendIoTime = CA_TIMEOUT;
    dd = 0;
    for (i=1 ; i < argc ; i++) {
      if (!strncmp("--",argv[i],2))
	dd = 1;
      else if ((!dd) && (!strcmp(argv[i],"-q") || !strcmp(argv[i],"-query")))
	*query = 1;
      else if ((!dd) && (!strcmp(argv[i],"-c") || !strcmp(argv[i],"-configure")))
	*configure = 1;
      else if ((!dd) && (!strcmp(argv[i],"-s") || !strcmp(argv[i],"-stop")))
	*stop = 1;
      else if ((!dd) && (!strcmp(argv[i],"-p") || 
	       !strcmp(argv[i],"-pend") ||
	       !strcmp(argv[i],"-pendIoTime"))) {
	i++;
	if (i >= argc) {
	  fprintf(stderr, "invalid -pendIoTime value (standardize)\n");
	  exit(1);
	}
	if ((sscanf(argv[i], "%lf", pendIoTime) != 1) ||
	    (*pendIoTime <= 0.0)) {
	  fprintf(stderr, "invalid -pendIoTime value (standardize)\n");
	  exit(1);
	}
      } else {
	if ((!dd) && (!strncmp("-",argv[i],1))) {
	  fprintf(stderr, "invalid option %s (standardize)\n", argv[i]);
	  cmderror++;
	} else {
	  strcpy(spec,argv[i]);
	}
      }
    }
  }

  if ((*query + *configure + *stop) > 1)
    cmderror++;
  if (strlen(spec)==0)
    cmderror++;

  if (cmderror) {
    usage(argv[0]);
    exit(1);
  }
  return(0);
}

/*************************************************************************
 * FUNCTION : usage()
 * PURPOSE  : Describes command line options available to client.
 * ARGS in  : char *name - name of executable program
 * ARGS out : none
 * GLOBAL   : nothing
 * RETURNS  : none
 ************************************************************************/  
void usage(char *name)
{
  fprintf(stderr,"usage: %s [-query -configure -stop]\n",name);
  fprintf(stderr,"       [-pendIoTime <seconds>] [--] spec-file\n");
  fprintf(stderr,"purpose:\tStandardizes the set of devices specified in\n");
  fprintf(stderr,"        \tspec-file. The spec-file must be in SDDS format.\n");
  fprintf(stderr,"        \tEach row in spec-file contains the device name,\n");
  fprintf(stderr,"        \trange over which to standardize, time over which\n");
  fprintf(stderr,"        \tto standardize, number of cycles, finishing setpoint,\n");
  fprintf(stderr,"        \tand whether to approach finish from above or below.\n");
  fprintf(stderr,"        \tEntering \"standardize myfile\" begins standardization.\n");
  fprintf(stderr,"options:\n");
  fprintf(stderr,"\t-query \n");
  fprintf(stderr,"\t\tPrint standardization state of devices in\n");
  fprintf(stderr,"\t\tspec-file (active/inactive).\n");
  fprintf(stderr,"\t-configure \n");
  fprintf(stderr,"\t\tConfigure subroutine records for\n");
  fprintf(stderr,"\t\tstandardization. Do not initiate.\n");
  fprintf(stderr,"\t-stop \n");
  fprintf(stderr,"\t\tHalt standardization of devices in spec-file.\n");
}

void sleep_us(unsigned int nusecs) {
   struct timeval tval;

   tval.tv_sec = nusecs / 1000000;
   tval.tv_usec = nusecs % 1000000;
   select(0, NULL, NULL, NULL, &tval);
 }
